/*      /\
 *     /  \		(C) Copyright 2008 Parliament Hill Computers Ltd.
 *     \  /		All rights reserved.
 *      \/
 *       .		Author: Alain Williams, September 2008
 *       .
 *        .		SCCS: @(#)KeepAlive.c	1.18 11/20/19 21:15:55
 *          .
 *
 * Send a ^q character every few minutes that it has been detected that the terminal
 * is idle. This should be run from a user's .profile.
 * Now options to send something other than ^q.
 *
 * This program will fork and the parent return on startup - this means that it
 * should not be run in background (with an & at the end of the command line).
 * This is so that:
 * * stdin can be used for idle time sensing
 * * this program is killed when the user logs out - OK I know that nohup is
 * 	something else, but ...
 *
 * The point about ^q is that it should be discarded by the terminal but count as
 * comms activity.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <stdarg.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <fcntl.h>
#include <pwd.h>
#include <ctype.h>
#include <getopt.h>

#define	MAX_TOUCH_STRING	1000	/* Max length of a string to -K option */

/* For character convertion */
#define	HexVal(x)	(isdigit(x) ? (x) - '0' : tolower(x) - 'a' + 10)
#define	OctVal(o)	((o) - '0')
#define	ToX(x)		("0123456789abcdef"[x])


#define	SECONDS_IN_MINUTE	60
#define	CONFIG_MAX_LINE_LENGTH	1024	// Longest line acceptable in a configuration file
#define	DEF_IDLE 		(5 * SECONDS_IN_MINUTE)	/* default idle - seconds */
#define	DEF_LOOP 		(1 * SECONDS_IN_MINUTE)	/* Loop time in - seconds */

char const* progname;	/* What this program is called - name of binary */
int	exit_c;		/* Exit status to return to O/S */
int	idleTime = DEF_IDLE;	/* How many seconds to be idle */
int	foreground;	/* Remain in the foreground */
int	loopTime = DEF_LOOP;	/* How long to wait between tests for idleness - seconds */
char	TouchChar = (char)('q' & 0x1f);	/* Character that gets output */
char*	TouchString = &TouchChar;	/* String to output - not NUL terminated */
int	TouchLen = 1;	/* Bytes in TouchString */
int	CheckParent;	/* Check life of Parent if this is true */
pid_t	Parent;		/* Our parent process if we are to check it */
int	CloseTerminal;	/* Check by filename */
char*	TerminalName;	/* Name of terminal connecte to stdout - only if CloseTerminal */
char*	Username;	/* Logged in username */
char*	HomeDirectory;	/* User's ~ - from password entry */
int	Verbose;	/* Be chatty to stderr */
int	Slept = 1;	/* True if we have slept & thus recorded time is wrong */

// **************** Command line processing ****************
extern	int	optind, opterr, optopt;
extern	char*	optarg;
// GNU extensions, useful but not portable:
// -W opt	If OptString contains 'W;' treat as --opt  This is special use of -W which is allowed by POSIX
// x::		x takes an optional argument
char const*	GetOptStr = ":cfi:I:k:K:NMpP:t:T:vx";	// Options (could dynamically build from long_options[] -- too much effort!)
// 2nd elem: 0 no arg, 1 mandatory arg, 2 optional arg. 3rd elem always set NULL
char const*	CmdLineOnly = "x";			// Command line only options
char const*	NotCmdLine  = "";			// Option not on command line

// Constants for long arguments - make > 255
#define	OPT_example_long_option	256	// --example_long_option

int	shortOptionOrVal;		// Used below & in read_opts() and OptName()
// Keep in alphabetic order:
const	struct option long_options[] = {
	{"check-parent",		0, &shortOptionOrVal, 'p'},
	{"check-pid",			1, &shortOptionOrVal, 'P'},
	{"close-terminal",		0, &shortOptionOrVal, 'c'},
	{"foreground",			0, &shortOptionOrVal, 'f'},
	{"help",			0, &shortOptionOrVal, 'x'},
	{"idle-minutes",		1, &shortOptionOrVal, 'i'},
	{"idle-seconds",		1, &shortOptionOrVal, 'I'},
	{"mobile",			0, &shortOptionOrVal, 'M'},
	{"send-char=c",			1, &shortOptionOrVal, 'k'},
	{"send-string",			1, &shortOptionOrVal, 'K'},
	{"send-nul",			0, &shortOptionOrVal, 'N'},
	{"test-interval-mins",		1, &shortOptionOrVal, 't'},
	{"test-interval-secs",		1, &shortOptionOrVal, 'T'},
	{"verbose",			0, &shortOptionOrVal, 'v'},
	{(char*)NULL,			0, (int*)NULL, 0}
};

char const* const	LongCmdLineOnly[] = {	// Array of long options only available on the command line
	"help",
	(char*)NULL
};

char const* const	LongNotCmdLine[] = {	// Array of long options not available on the command line
	(char*)NULL
};


/* Usage message.
 */
char const* const	usagem[] = {
	"Keep an idle terminal session alive by sending ^q characters",
	"Usage: %s [options]",
	"Options:",
	"-c --close-terminal",
	"	Close the terminal before going into background - may allow HUP to be sent",
	"-f --foreground",
	"	Remain in the foreground (debug option)",
	"-i n --idle-minutes=n",
	"	Idle time in minutes before ^q sent",
	"-I n --idle-seconds=n",
	"	Idle time in seconds before ^q sent",
	"-k c --send-char=c",
	"	Keepalive character to output, either 1 character or a decimal integer or something like ^q",
	"-K str --send-string=str",
	"	String to output, may contain chars, ^c, \\xdd, \\ooo",
	"-N --send-nul",
	"	send NUL bytes and other options, same as: -k ^@ -p -c -i 1",
	"-M --mobile",
	"	suitable for maintaining a Mobile 'phone link. Same as: -k ^@ -p -c -I 10",
	"-p --check-parent",
	"	check on Parent process (hopefully the login shell), die if it dies",
	"-P n --check-pid=n",
	"	process id to check when -p given, when parent is not appropriate",
	"-t n --test-interval-mins=n",
	"	Test time in minutes between checks that the terminal is idle",
	"-T n --test-interval-secs=n",
	"	Test time in seconds between checks that the terminal is idle",
	"-v --verbose",
	"	Verbose - send log messages to stderr",
	"-x --help",
	"	eXplain",
	"Author: Alain D D Williams <addw@phcomp.co.uk> Sept 2008, licenced GPL 2",
	(char*)NULL	/* Must end the list */
};

/* Functions defined in this module, in alphabetic order */
void			die(char* fmt, ...) __attribute__ ((noreturn));
void			GetTerminalName(void);
time_t			GetTime(void);
void			GetUserName(void);
void			KeepAlive(void);
void			logMsg(char* fmt, ...);
const char*		OptName(int c, char const* const argv[]);
int			ProcArg(int c, char const* opt, char const* where);
void			ProcConfFile(char* ConfFile);
int			ProcConfLine(char* cp, char const* Errm);
void			ReadOptions(int argc, char const* const argv[]);
void			read_opts(int argc, char const* const argv[]);
void			ReadTouchString(char* optarg);
void			SetupTerminal(void);
char const* const*	strInArray(char const* const* strarray, const char* find);
char*			strmake(char const* Arg, char const* where);
void			usage(int exitc);

int	main(int argc, char const* const argv[])
{
	Parent = getppid();	/* Set default */
	progname = argv[0];

	GetUserName();	// need ~ ($HOME)

	ReadOptions(argc, argv);

	/* Sanity check */
	if(loopTime > idleTime)
		loopTime = idleTime;


	SetupTerminal();

	argv += optind;
	if(*argv)
		fprintf(stderr, "%s: Extra command line argument(s) ignored\n", progname);

	if( ! foreground)	/* Don't fork in debug mode */
		switch(fork()){
		case -1:
			die("Failed to create background process: %s", strerror(errno));
		case 0:	/* Child - continue to execute */
			break;
		default:/* Parent - exit and allow the .profile to continue */
			exit(0);
		}

	logMsg("Starting monitor, pid %d, idle time %d secs, loop %d secs, char %2.2x", (int)getpid(), idleTime, loopTime, TouchChar);

	KeepAlive();

	return(exit_c);
}

/* Read options/arguments from various places:
 * /etc/KeepAlive.rc
 * ~/.KeepAlive.rc
 * command line
 *
 * If something print a message and exit
 */
void	ReadOptions(int argc, char const* const argv[])
{
	ProcConfFile("/etc/KeepAlive.rc");

	// If we know the home directory:
	if(*HomeDirectory) {
		char*	home;
		char*	homerc = "/.KeepAlive.rc";

		if( !(home = (char*)malloc(strlen(HomeDirectory) + strlen(homerc) + 1)))
			die("Failed to allocate string for ~/.KeepAlive.rc");
		strcat(strcpy(home, HomeDirectory), homerc);
		ProcConfFile(home);
		free(home);
	}

	// Command line:
	read_opts(argc, argv);
}

/* Interpret '-' arguments -options.
 * If there is an error in the arguments, print the options list and exit.
 */
void	read_opts(int argc, char const* const argv[])
{
	register int	c;
	register int	error;
	int		i;

	error = 0;

	opterr = 1;	// Stop getopt_long printing error messages - we do it

	// Command line arguments:
	while((c = getopt_long(argc, (char* const*)argv, GetOptStr, long_options, &i)) != -1) {
		// This is a bit convoluted. Some short and long options are equivalent.
		// The simple way of managing this would be to set 'flag' (3rd member in each long_options)
		// to (char*)NULL - this would work.
		// However: when there is an error it is nice to be able to tell the user showing the exact
		// option that they used, ie not the long one if they gave the short one.

		// The only way to distinguish which was used is to set the 'flag', then the value of 'val'
		// is stored there. getopt_long() then returns 0 when a long option is recognised.
		// In this program 'flag' always points to shortOptionOrVal.

		// If it is a long only option the value will be > 255

		// See also OptName() complication about this

		const char*	opt_as_typed = OptName(c, argv);

		if(c == 0)	// Get the short (-x) value - this is what we test for
			c = shortOptionOrVal;

		if((c <= 255 && strchr(NotCmdLine, c)) || (c > 255 && strInArray(LongNotCmdLine, long_options[i].name))) {
			fprintf(stderr, "%s option '%s' is not available on the command line\n", progname, opt_as_typed);
			error = 1;
		} else
			error |= ProcArg(c, opt_as_typed, "command line");
	}

	if(error)
		usage(2);
}

/* Read and set options from a config (.rc) file.
 * It is not an error if the file does not exist.
 *
 * If there is an error, stop the program at the end of the file.
 */
void	ProcConfFile(char* ConfFile)
{
	int	error = 0;
	FILE*	cf;
	int	LineNo = 0;
	char	buf[CONFIG_MAX_LINE_LENGTH + 1];
	char	Errm[200];

	if(!(cf = fopen(ConfFile, "r")))
		return;

	while(fgets(buf, sizeof(buf), cf)) {
		char*	cp;

		LineNo++;
		snprintf(Errm, sizeof(Errm), "config file '%s' line %d", ConfFile, LineNo);

		if((cp = strchr(buf, '\n')) != (char*)NULL)
			*cp = '\0';

		// Be nice to idiots who edit files on a MS Windows machine
		if((cp = strrchr(buf, '\r')) && cp[1] == '\0')
			*cp = '\0';

		// Trim: forwards & back
		cp = strchr(buf, '\0') - 1;
		while(cp >= buf && (*cp == ' ' || *cp == '\t'))
			*cp-- = '\0';
		cp = buf;
		while(*cp == ' ' || *cp == '\t')
			cp++;

		// Skip empty lines & comments
		if(*cp == '\0' || *cp == '#')
			continue;

		error |= ProcConfLine(cp, Errm);

	}

	fclose(cf);

	if(error)
		usage(2);
}

/* Process a line out of the conf file - cp
 * Errm is a string saying something about where the error was found (eg line #)
 * Cope with options of the form '-c' and '--option' and 'option'
 *
 * Return true if error.
 *
 * Simplified & borrowed from SeekSpeed.c
 */
int	ProcConfLine(char* cp, char const* Errm)
{
	char*	os;
	int	opt;
	int	OptHasArg;
	char*	optStr;
	char	endCh;
	char	optBuf[4];
	struct option const*	op;

	if(*cp == '-' && cp[1] != '-') {
		// Single letter option, eg: -x
			
		optBuf[0] = '-';	// For error message
		optBuf[1] = cp[1];
		optBuf[2] = '\0';
		optStr = optBuf;

		// Search the option string:
		if(*++cp == '\0' || *cp == ':' || !(os = strchr(GetOptStr, *cp))) {
			fprintf(stderr, "%s unrecognised option: '%s'\n", Errm, optStr);
			return(1);
		}

		if(strchr(CmdLineOnly, *cp)) {
			fprintf(stderr, "%s option only allowed on command line: '%s'\n", Errm, optStr);
			return(1);
		}

		OptHasArg = os[1] == ':';
		opt = *cp++;

		endCh = ' ';	// Set to anything but '=' so that eating of spaces will happen
	} else {
		// Multiple letter option names
		char*	optionName;

		optStr = cp;
		// It can be '--something' or just 'something'; treat them as the same
		if(*cp == '-' && cp[1] == '-') {
			cp += 2;

			// Expect a word after the --
			if(*cp == '\0' || *cp == ' ' || *cp == '\t') {
				fprintf(stderr, "%s found '--' not followed by an option\n", Errm);
				return(1);
			}
		}
		// Check if it is --something
		optionName = cp;

		// Find the end of the option & NUL terminate
		while(*++cp != '\0' && *cp != ' ' && *cp != '\t' && *cp != '=')
				;
		endCh = *cp;
		*cp = '\0';

		// Look it up
		for(op = long_options; ; op++) {
			if( !op->name) {
				fprintf(stderr, "%s unrecognised option: '%s'\n", Errm, optStr);
				return(1);
			}

			if( !strcmp(op->name, optionName))
				break;		// Found it!
		}

		// What do we know about it ?
		opt = op->val;			// Value to switch on
		OptHasArg = op->has_arg == 1;	// Must this have an argument ? Optional args not handled

		// Skip the '\0' to next char on the line - unless we already reached end of line
		if(endCh != '\0')
			cp++;

		// If the ending character was white space, span foward on white space
		// eat any '=' that we may find. This is to allow something like:
		// option-name  =  option-value
		if(endCh == ' ' || endCh == '\t') {
			while(*cp == ' ' || *cp == '\t')
				cp++;
			if(*cp == '=')
				cp++;
			// Any spaces after the '=' are eaten below
		}

		// Some are for the command line only:
		if(strInArray(LongCmdLineOnly, optStr + 2)) {
			fprintf(stderr, "%s option only allowed on the command line: '%s'\n", Errm, optStr);
			return(1);
		}
	}

	// See if there is anything else on the line, span spaces.
	while(*cp == ' ' || *cp == '\t')
		cp++;

	if(OptHasArg) {
		if( !*cp) {
			fprintf(stderr, "%s option needs an argument: '%s'\n", Errm, optStr);
			return(1);
		}

		// If the argument starts and ends with a quote character, remove it:
		if(*cp == '\'' || *cp == '"') {
			char*	endch = cp + strlen(cp) - 1;

			if(*endch == *cp && endch != cp) {
				cp++;
				*endch-- = '\0';
			}
		}

		// The argument is the remaining characters on the line.
		optarg = cp;

	} else if( *cp != '\0') {
		fprintf(stderr, "%s option should not have argument: '%s%c%s'\n", Errm, optStr, endCh, cp);
		return(1);
	}

	return(ProcArg(opt, optStr, Errm));
}

/* Process an argument, be that from the command line of a .rc file:
 * c	The argument letter or something like OPT_example_long_option
 * opt	The string that makes the option, eg '-x' '--help' 'help' --- used in error messages
 * where Where the argument came from - for error messages
 *
 * **** WARNING ****
 * If the option has an argument it will be in global: optarg - beware, this might be over written,
 * so if the string value is to be kept a copy must be made.
 *
 * Return true if error
 * This (& elsewhere) can leak memory - but small amounts - not important.
 */
int	ProcArg(int c, char const* opt, char const* where)
{
	int	error = 0;
	long	tl;	/* temp long */
	char*	te;	/* temp end pointer */

	// printf("ProcArg(int %c, opt '%s', char const* '%s') optarg='%s'\n", c, opt, where, optarg);

	switch(c) {
	case 'c':	/* Close tty */
		CloseTerminal = 1;
		break;
	case 'f':	/* Foreground */
		foreground = 1;
		break;
	case 'i':	/* Idle time */
		tl = strtol(optarg, &te, 10);
		if(te == optarg || *te != '\0')
			die("argument to %s is not numeric: '%s'", opt, optarg);
		if(tl <= 0)
			die("idle interval must be positive: '%s %s'", opt, optarg);
		idleTime = (int)tl * SECONDS_IN_MINUTE;
		break;
	case 'I':	/* Idle time -- seconds */
		tl = strtol(optarg, &te, 10);
		if(te == optarg || *te != '\0')
			die("argument to %s is not numeric: '%s'", opt, optarg);
		if(tl <= 0)
			die("idle interval must be positive: '%s %s'", opt, optarg);
		idleTime = (int)tl;
		break;
	case 'k':	/* Keepalive character to output */
		/* Just one character ? */
		if(strlen(optarg) == 1 && !isdigit(*optarg)) {
			TouchChar = *optarg;
			break;
		}

		/* Something like ^q ? */
		if(strlen(optarg) == 2 && *optarg == '^') {
			TouchChar = optarg[1] & 0x1f;
			break;
		}

		/* Interpret as decimal integer */
		tl = strtol(optarg, &te, 10);
		if(te == optarg || *te != '\0')
			die("argument it %s is not numeric: '%s'", opt, optarg);
		if(tl < 0 || tl > 255)
			die("Idle interval must be between 0 & 255: '%s %s'", opt, optarg);
		TouchChar = (char)tl;

		/* (re)set to this character */
		if(TouchString != &TouchChar)	/* In case anything set from a previous string */
			free(TouchString);
		TouchString = &TouchChar;
		TouchLen = 1;
		break;
	case 'K':	/* Keepalive string to output */
		ReadTouchString(optarg);
		break;
	case 'N':	/* send NUL bytes and other options, same as: -k ^@ -p -c -i 1
			 * The point is that I found myself using this combination a lot.
			 */
		TouchChar = '\0';
		TouchString = &TouchChar;
		TouchLen = 1;
		CheckParent = 1;
		CloseTerminal = 1;
		idleTime = 1 * SECONDS_IN_MINUTE;
		break;
	case 'M':	/* send NUL bytes and other options, same as: -k ^@ -p -c -I 10
			 * -N is not frequent enough for a link over a 3G Mobile telephone.
			 */
		TouchChar = '\0';
		TouchString = &TouchChar;
		TouchLen = 1;
		CheckParent = 1;
		CloseTerminal = 1;
		idleTime = 10;
		break;
	case 'p':	/* Check parent process */
		CheckParent = 1;
		break;
	case 'P':	/* Parent process ID */
		tl = strtol(optarg, &te, 10);
		if(te == optarg || *te != '\0')
			die("argument to %s is not numeric: '%s'", opt, optarg);
		if(tl <= 0)
			die("Parent process id must be positive: '%s %s'", opt, optarg);
		Parent = (pid_t)tl;
		break;
	case 't':	/* Test time */
		tl = strtol(optarg, &te, 10);
		if(te == optarg || *te != '\0')
			die("argument to %s is not numeric: '%s'", opt, optarg);
		if(tl <= 0)
			die("test interval must be positive: '%s %s'", opt, optarg);
		loopTime = (int)tl * SECONDS_IN_MINUTE;
		break;
	case 'T':	/* Test time -- seconds */
		tl = strtol(optarg, &te, 10);
		if(te == optarg || *te != '\0')
			die("argument to %s is not numeric: '%s'", opt, optarg);
		if(tl <= 0)
			die("test interval must be positive: '%s %s'", opt, optarg);
		loopTime = (int)tl;
		break;
	case 'x':	/* eXplain */
		usage(0);
		/*NOTREACHED*/
	case 'v':	/* Verbose */
		Verbose = 1;
		break;
	case ':':
		die("Argument missing to option '%s'", opt);
		error = 1;
		break;
	case '?':
		die("Unknown option '%s'", opt);
		error = 1;
		break;
	default:
		die("%s Internal error, unrecognised option '%s'", where, opt);
	}

	return(error);
}

/* Return a string to a representation of the option with code c
 * The option may be a short or long one. The long name is returned by preference.
 * The string is in a static area.
 * If c is 0 then the option will have been a long one, shortOptionOrVal
 * This is only used when processing command line options.
 */
const char*	OptName(int c, char const* const argv[])
{
	static	char	opt[100] = "?";
	struct option*	op;

	// Dig through options to convert c to the long name.
	// Try this first, options have a short and long form, if we are here
	// we don't know which was given since the value for the long form is the short one.
	// Give the long one in preference since it is more meaningful.
	if(c == 0) {
		for(op = (struct option*)long_options; op->name; op++)
			if(op->val == shortOptionOrVal) {
				snprintf(opt, sizeof(opt), "--%s", op->name);
				return(opt);
			}

		// Should not get here
		snprintf(opt, sizeof(opt), "Unknown opt --%d (program error)", shortOptionOrVal);
		return(opt);
	}

	// printf("optind=%d '%s'\n", optind, argv[optind - 1]);

	// Unknown option, the character that is not known is in optopt.
	// Ditto if an argument is missing
	if(c == '?' || c == ':') {
		// Check the current option being processed. If it starts -- return that
		// Otherwise we get the single character option name - which will confuse
		if(argv[optind - 1][0] == '-' && argv[optind - 1][1] == '-')
			return(argv[optind - 1]);

		c = optopt;
	}

	// It was a short option
	if(c <= 255) {
		// Try to print nicely
		if(iscntrl(c))
			snprintf(opt, sizeof(opt), "-^%c", c | 0x40);
		else
			snprintf(opt, sizeof(opt), "-%c", c);
	} else {
		// Should not happen
		snprintf(opt, sizeof(opt), "--bad_long=%d", c);
	}

	return(opt);
}



/* Read a string and set TouchString & TouchLen
 * The area read into is malloced
 * Interpret: ^c, \xdd, \ooo
 */
void	ReadTouchString(char* optarg)
{
	if(TouchString != &TouchChar)	/* In case anything set from a previous string */
		free(TouchString);

	/* We need somewhere to keep this, malloc something.
	 * A good length would be the input string - it won't be longer than that.
	 * Put a max limit on that - just in case.
	 */
	if((TouchLen = strlen(optarg)) >= MAX_TOUCH_STRING)
		die("String to -K option is too long (%d), max is %d", TouchLen, MAX_TOUCH_STRING);

	if(TouchLen == 0)
		die("Empty string to -K option is not allowed");

	if( ! (TouchString = (char*)malloc(TouchLen)))
		die("Failed to malloc %d bytes for -K string", TouchLen);

	TouchLen = 0;

	while(*optarg)
		switch(*optarg) {
		case '^':	/* Control of the next character */
			if(*++optarg == '\0')
				die("Bad argument to -K, ^ cannot be at end of string");
			TouchString[TouchLen++] = *optarg++ & 0x1f;
			break;

		case '\\':	/* Hex or Octal char */
			if(*++optarg == '\0')
				die("Bad argument to -K, \\ cannot be at end of string");
			if(*optarg == 'x' || *optarg == 'X') {
				if(strlen(++optarg) < 2 || !isxdigit(*optarg) || !isxdigit(optarg[1]))
					die("Bad argument to -K, \\x not followed by 2 hex digits");

				TouchString[TouchLen++] = (char)((HexVal(*optarg) << 4) | HexVal(optarg[1]));
				optarg += 2;
				break;
			}

			/* Up to 3 octal digits */
			if(isdigit(*optarg) && *optarg != '8' && *optarg != '9') {
				TouchString[TouchLen] = (char)OctVal(*optarg++);
				if(isdigit(*optarg) && *optarg != '8' && *optarg != '9') {
					TouchString[TouchLen] = (char)(TouchString[TouchLen] << 3 | OctVal(*optarg++));
					if(isdigit(*optarg) && *optarg != '8' && *optarg != '9')
						TouchString[TouchLen] = (char)(TouchString[TouchLen] << 3 | OctVal(*optarg++));
				}
				TouchLen++;
				break;
			}

			die("Bad argument to -K, \\ must be followed by 'x' or an octal digit");
			break;

		default:	/* All else just put into the string */
			TouchString[TouchLen++] = *optarg++;
		}

	/* Show what we did ? */
	if(Verbose) {
		int	i;
		char*	hex;	/* TouchString in hex */
		char*	play;

		if( ! (hex = (char*)malloc(i = TouchLen * 3)))
			die("Failed to malloc %d bytes for -K string debug", i);

		play = hex;
		i = 0;
		while(i < TouchLen) {
			*play++ = ToX(TouchString[i] >> 4 & 15);
			*play++ = ToX(TouchString[i++]    & 15);
			*play++ = ' ';
		}
		play[-1] = '\0';

		logMsg("Converted -K string as 0x'%s'", hex);
		free(hex);
	}
}

/* Output a usage message and end with exit status exitc.
 */
void	usage(int exitc)
{
	register char const* const *	mes;
	
	for(mes = usagem; *mes != (char*)NULL; mes++) {
		(void) fprintf(stderr, *mes, progname);
		fputc('\n', stderr);
	}

	fprintf(stderr, "Default idle minutes is %d, minutes between tests %d\n", DEF_IDLE, DEF_IDLE);

	exit(exitc);
	/*NOTREACHED*/
}

/* Print an error and kill the program.
 * Arguments as printf()
 */
void	die(char* fmt, ...)
{
	va_list ap;

	fprintf(stderr, "%s: Error: ", progname);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fprintf(stderr, "\n");

	exit(2);
}

/* Find a string in an array of strings.
 * strarray - NULL terminated array - haystack
 * find     - what to look for
 * Return: if found the address of member in strarray, otherwise NULL
 */
char const* const*	strInArray(char const* const* strarray, const char* find)
{
	for(; *strarray; strarray++)
		if( !strcmp(*strarray, find))
			return(strarray);

	return((char const *const *)NULL);
}

/* Return an allocated copy of the argument string Arg.
 * If the allocation fails, kill the program with where as part of the error message.
 */
char*	strmake(char const* Arg, char const* where)
{
	char*	s;

	if((s = strdup(Arg)) == (char*)NULL)
		return(s);

	die("%s out of memory in string duplicate", where);
}

/* Return the time.
 * Only call time() if Slept is true - save on syscalls.
 */
time_t	GetTime(void)
{
	static	time_t	Now;

	if(Slept) {
		Now = time((time_t*)NULL);
		Slept = 0;
	}

	return(Now);
}

/* Log a message to stderr, prefix with the time, stamp every day.
 * Only log if Verbose.
 * Arguments as printf()
 */
void	logMsg(char* fmt, ...)
{
	va_list 		ap;
	static	struct tm	LastTM;	/* Time we were last here */
        static	struct tm*	NowTM;	/* Time now */
	time_t			now;

	if( ! Verbose)
                return;

	GetTerminalName();

	now = GetTime();
	NowTM = localtime(&now);

        /* Print date & tty name if changed since we were last here */
        if(NowTM->tm_yday != LastTM.tm_yday) {
		fprintf(stderr, "%4.4d/%2.2d/%2.2d %2.2d:%2.2d:%2.2d %s Monitoring %s, user %s\n",
			NowTM->tm_year + 1900, NowTM->tm_mon + 1, NowTM->tm_mday,
			NowTM->tm_hour, NowTM->tm_min, NowTM->tm_sec,
			progname, TerminalName, Username);
	}

	/* Always show current time */
	fprintf(stderr, "%2.2d:%2.2d:%2.2d ", NowTM->tm_hour, NowTM->tm_min, NowTM->tm_sec);

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fprintf(stderr, "\n");

	fflush(stderr);	/* May not write for a long time */

	memcpy((void*)&LastTM, (void*)NowTM, sizeof(LastTM));	/* Set last time */
}

/* Get the name in TerminalName - if not already got.
 */
void	GetTerminalName(void)
{
	if(TerminalName)	/* Already done ? */
		return;

	/* What terminal are we connected to ? */
	if( ! (TerminalName = ttyname(1)))
		die("can't determine where stdout is connected to: %s", strerror(errno));
}

/* Put something in Username - if not already there.
 * Also get ~ ($HOME) into HomeDirectory -- "" if not known
 */
void	GetUserName(void)
{
	uid_t			uid;
	struct passwd*		pwuid;
	char*			home;
	char*			errm = "Program initialisation - GetUserName";

	if(Username)		/* Already done ? */
		return;

	/* Get username (for logging) */
	uid = getuid();	
	if((pwuid = getpwuid(uid)) && pwuid->pw_name)
		Username = strmake(pwuid->pw_name, errm);
	else {
		char	Buf[40];
		snprintf(Buf, sizeof(Buf) - 1, "uid %d", (int)uid);
		Username = strmake(Buf, errm);
	}

	/* Get ~ ($HOME) for defaults file */
	home = (pwuid && pwuid->pw_dir) ? pwuid->pw_dir : "";
	HomeDirectory = strmake(home, errm);
}

/* If CloseTerminal get the name in TerminalName and close the terminal.
 */
void	SetupTerminal(void)
{
        char*	tmp;

	if( ! CloseTerminal && ! Verbose)	/* If verbose need the ttyname */
		return;

	GetTerminalName();

	if( ! CloseTerminal)
		return;

	if( ! (TerminalName = strdup(TerminalName)))
		die("out of memory!! can't strdup TerminalName");

	/* stderr may not be connected to a terminal - or the same one */
        if( (tmp = ttyname(2)) && !strcmp(TerminalName, tmp))
		close(2);	/* May well screw up die() */

	/* Close everything that we have on the tty */
	close(1);
	close(0);
}

/* Monitor the terminal and send ^q characters
 * This will prob end when the user logs out.
 */
void	KeepAlive(void)
{
	struct stat sb;
	time_t	now;
        int	ttyfd;				/* Where we have the terminal open */
	int	c_output;			/* Count of chars output */

	while(1) {
		sleep(loopTime);
		Slept = 1;

		/* If wanted check that our parent still exists.
		 * Generating an error message is somewhat superfluous.
		 */
		if(CheckParent && kill(Parent, 0) == -1) {
			if(errno == ESRCH)
				die("it looks as if the user has logged out");
			else
				die("sending signal to parent failed: %s", strerror(errno));
		}

		/* We may/may-not have the terminal open */
		if(CloseTerminal) {
			if(stat(TerminalName, &sb) == -1)
				die("stat check on the terminal %s failed: %s", TerminalName, strerror(errno));
		} else {
			if(fstat(1, &sb) == -1)
				die("stat check on the terminal fd failed: %s", strerror(errno));
		}

		now = GetTime();

		logMsg("Terminal accessed %ld (%ld ago), updated %ld (%ld ago), now is %ld", (long)sb.st_atime,
			(long)now - (long)sb.st_atime, (long)sb.st_mtime, (long)now - (long)sb.st_mtime, (long)now);

		now -= idleTime;	/* Make comparison easier */

		if(sb.st_atime > now || sb.st_mtime > now)
			continue;	/* There has been recent activity */

		if(CloseTerminal) {
                        if((ttyfd = open(TerminalName, O_WRONLY)) == -1)
				die("can't open terminal '%s' as: %s", TerminalName, strerror(errno));
                } else
			ttyfd = 1;

		logMsg("Writing %d bytes to terminal", TouchLen);

		c_output = 0;
		while(c_output < TouchLen) {
			int	i;

			/* This can fail if we don't get killed by HUP properly */
			if((i = write(ttyfd, TouchString + c_output, TouchLen - c_output)) == -1) {
				if(errno == EINTR)
					continue;

				die("Write of keepAlive string failed: %s", strerror(errno));
			}

			if(i == 0)
				die("Write of keepAlive string wrote zero ? Aborting");

			c_output += i;

			logMsg("Wrote %d, total this time %d", i, c_output);
		}

		if(CloseTerminal)	/* Don't keep it open if asked to close */
                        close(ttyfd);
	}
}

/* end */
